#include <stdio.h>
#include <string.h>
#include <libsx.h>
#include "maze.h"

/**** Constants ****/
#define BUFLEN 1024

/**** Variables ****/
/* Maze Map */
int HSize = 0;
int VSize = 0;
int **Maze = NULL;
int StartX = -1;
int StartY = -1;
int GoalX = -1;
int GoalY = -1;

/* Location, Direction */
int PosX;
int PosY;
int Dir = 3;
int DirX[4] = { 0, 1, 0, -1};	/* Up, Right, Down, Left */
int DirY[4] = {-1, 0, 1,  0};
ViewInfo View;
int **SysmtemMap;
int Stat = STAT_STEP1;
int Step1Count = 0;
int Step2Count = 0;
int AutoRun = 0;

/* X11 */
int ViewWidth;
int ViewHight;
int MapWidth;
int MapHight;
char StepString[256];

Widget QuitButton;
Widget TurnLeftButton;
Widget ForwardButton;
Widget TurnRightButton;
Widget StatLabel;

Widget Step1Label;
Widget Step1Button;
Widget Auto1Button;
Widget Count1Label;

Widget SearchLabel;
Widget SearchButton;

Widget Step2Label;
Widget Step2Button;
Widget Auto2Button;
Widget Count2Label;

Widget ViewScreen;
Widget MapScreen;

/**** Functions ****/
/*
  Load(char *filename): Load maze from file
    File format:
      hsize SIZE
      vsize SIZE
      begin
        :
      end

      S: start
      G: goarl
       : route
      ?: wall
*/

void Load(char *filename)
{
  FILE *fp;
  char buf[BUFLEN];
  int lines = 0;
  int stat = 0;
  int num = 0;
  int i, j, l;

  if ((fp = fopen(filename, "r")) == NULL) {
    fprintf(stderr, "Cannot open file '%s'.\n", filename);
    exit(1);
  }
  while (fgets(buf, BUFLEN, fp) != NULL) {
    lines++;
    /*    fprintf(stderr, "%d:%d:%s", stat, lines, buf); */
    if (strncmp(buf, "hsize", 5) == 0) {
      if (sscanf(buf, "hsize %d", &HSize) != 1) {
	fprintf(stderr, "Bad hsize '%s' at %d\n", buf, lines);
	exit(1);
      }
      if (HSize >= BUFLEN) {
	fprintf(stderr, "Too large hsize %d (max: %d)\n", HSize, BUFLEN - 1);
	exit(1);
      }
      if (HSize < 1) {
	fprintf(stderr, "hsize must be larger than 0\n");
	exit(1);
      }
      if (VSize > 0)
	stat = 1;
    }
    else if (strncmp(buf, "vsize", 5) == 0) {
      if (sscanf(buf, "vsize %d", &VSize) != 1) {
	fprintf(stderr, "Bad vsize '%s' at %d\n", buf, lines);
	exit(1);
      }
      if (VSize < 1) {
	fprintf(stderr, "vsize must be larger than 0\n");
	exit(1);
      }
      if (HSize > 0)
	stat = 1;
    }
    else if (strncmp(buf, "begin", 5) == 0) {
      if (stat < 1) {
	fprintf(stderr, "Define hsize and vsize before 'begin'\n");
	exit(1);
      }
      stat = 2;
      if (Maze == NULL) {
	if ((Maze = (int **)calloc(HSize, sizeof(int *))) == NULL) {
	  fprintf(stderr, "Cannot allocate memory for Maze(%d)\n", HSize);
	  exit(1);
	}
	if ((SysmtemMap = (int **)calloc(HSize, sizeof(int *))) == NULL) {
	  fprintf(stderr, "Cannot allocate memory for SystemMap(%d)\n", HSize);
	  exit(1);
	}
	for (i = 0; i < HSize; i++) {
	  if ((Maze[i] = (int *)calloc(VSize, sizeof(int))) == NULL) {
	    fprintf(stderr, "Cannot allocate memory for Maze[%d](%d)\n",
		    i, VSize);
	    exit(1);
	  }
	  if ((SysmtemMap[i] = (int *)calloc(VSize, sizeof(int))) == NULL) {
	    fprintf(stderr, "Cannot allocate memory for SystemMap[%d](%d)\n",
		    i, VSize);
	    exit(1);
	  }
	  for (j = 0; j < VSize; j++)
	    SysmtemMap[i][j] = WALL;
	}
      }
    }
    else if (strncmp(buf, "end", 3) == 0) {
      if (num < VSize) {
	fprintf(stderr, "No enough lines\n");
	VSize = num;
      }
      if (StartX < 0) {
	fprintf(stderr, "No start\n");
	exit(1);
      }
      if (GoalX < 0) {
	fprintf(stderr, "No goal\n");
	exit(1);
      }
      stat = 3;
    }
    else {
      if (stat == 2) {
	if (num >= VSize) {
	  fprintf(stderr, "Too many data at %d: %s\n", lines, buf);
	}
	else {
	  l = strlen(buf);
	  if (buf[l - 1] == 0x0a) {
	    l--;
	    buf[l] = 0;
	  }
	  if (l > HSize) {
	    fprintf(stderr, "Too long line at %d (%d): %s\n", lines, l, buf);
	    l = HSize;
	  }
	  else if (l < HSize) {
	    fprintf(stderr, "Too short line at %d(%d): %s\n", lines, l, buf);
	  }
	  for (i = 0; i < l; i++) {
	    if (buf[i] == ' ')
	      Maze[i][num] = ROUTE;
	    else if (buf[i] == 'S') {
	      Maze[i][num] = START;
	      PosX = StartX = i;
	      PosY = StartY = num;
	      SysmtemMap[StartX][StartY] = START;
	    }
	    else if (buf[i] == 'G') {
	      Maze[i][num] = GOAL;
	      GoalX = i;
	      GoalY = num;
	    }
	    else
	      Maze[i][num] = WALL;
	  }
	  for (; i < HSize; i++) {
	    Maze[i][num] = -1;
	  }
	  num++;
	}
      }
      else {
	fprintf(stderr, "Syntax error at %d: %s\n", lines, buf);
      }
    }
  }
  fclose(fp);
}

/* Show Maze */
void Show(int **Maze)
{
  int x, y;

  fprintf(stderr, "     ");
  for (x = 0; x < HSize; x++) {
    fprintf(stderr, "%d", (x + 1) % 10);
  }
  fprintf(stderr, "\n");
  for (y = 0; y < VSize; y++) {
    fprintf(stderr, "%3d: ", y + 1);
    for (x = 0; x < HSize; x++) {
      switch (Maze[x][y]) {
      case UNKNOWN:
	fprintf(stderr, "?");
	break;
      case GOAL:
	fprintf(stderr, "G");
	break;
      case START:
	fprintf(stderr, "S");
	break;
      case WALL:
	fprintf(stderr, "#");
	break;
      case ROUTE:
	fprintf(stderr, " ");
	break;
      case 1:
	fprintf(stderr, "+");
	break;
      case 2:
	fprintf(stderr, ".");
	break;
      default:
	fprintf(stderr, "*");
	break;
      }
    }
    fprintf(stderr, "\n");
  }
}

/* Get ViewInfo */
void GetViewInfo(ViewInfo *info, int **Maze, int PosX, int PosY, int Dir)
{
  int d;

  /* Left */
  d = Dir - 1;
  if (d < 0)
    d += 4;
  info->left = Maze[PosX + DirX[d]][PosY + DirY[d]];

  info->front = Maze[PosX + DirX[Dir]][PosY + DirY[Dir]];

  /* Right */
  d = Dir + 1;
  if (3 < d)
    d -= 4;
  info->right = Maze[PosX + DirX[d]][PosY + DirY[d]];

  /* Back */
  d = Dir + 2;
  if (3 < d)
    d -= 4;
  info->back = Maze[PosX + DirX[d]][PosY + DirY[d]];

  info->current = Maze[PosX][PosY];

  info->dir = Dir;
  info->x = PosX;
  info->y = PosY;
}

/* Write ViewInfo into Maze Map */
void SetViewInfo(ViewInfo *info, int **Maze)
{
  int d;
  int PosX, PosY, Dir;

  PosX = info->x;
  PosY = info->y;
  Dir = info->dir;

  /* Left */
  d = Dir - 1;
  if (d < 0)
    d += 4;
  Maze[PosX + DirX[d]][PosY + DirY[d]] = info->left;

  /* Front */
  Maze[PosX + DirX[Dir]][PosY + DirY[Dir]] = info->front;

  /* Right */
  d = Dir + 1;
  if (3 < d)
    d -= 4;
  Maze[PosX + DirX[d]][PosY + DirY[d]] = info->right;

  /* Back */
  d = Dir + 2;
  if (3 < d)
    d -= 4;
  Maze[PosX + DirX[d]][PosY + DirY[d]] = info->back;

  /* Current Location */
  Maze[PosX][PosY] = info->current;
}

/* Get New ViewInfo after Next Action */
/*
	return 0 if success
	return -1 if failed
*/
int GetNextViewInfo(ViewInfo *info, int **Maze, int next)
{
  int x, y, dir, newdir;

  x = info->x;
  y = info->y;
  dir = info->dir;

  switch(next) {
  case GO_LEFT:
    if (info->left != WALL) {
      if (--dir < 0)
	dir += 4;
      x += DirX[dir];
      y += DirY[dir];
      GetViewInfo(info, Maze, x, y, dir);
      return 0;
    }
    else
      return -1;
    break;
  case TURN_LEFT:
    if (--dir < 0)
      dir += 4;
    GetViewInfo(info, Maze, x, y, dir);
    return 0;
    break;
  case GO_FORWARD:
    if (info->front != WALL) {
      x += DirX[dir];
      y += DirY[dir];
      GetViewInfo(info, Maze, x, y, dir);
      return 0;
    }
    else
      return -1;
    break;
  case TURN_RIGHT:
    if (++dir > 3)
      dir -= 4;
    GetViewInfo(info, Maze, x, y, dir);
    return 0;
    break;
  case GO_RIGHT:
    if (info->right != WALL) {
      if (++dir > 3)
	dir -= 4;
      x += DirX[dir];
      y += DirY[dir];
      GetViewInfo(info, Maze, x, y, dir);
      return 0;
    }
    else
      return -1;
    break;
  case GO_BACK:
    if (info->back != WALL) {
      dir += 2;
      if (dir > 3)
	dir -= 4;
      x += DirX[dir];
      y += DirY[dir];
      GetViewInfo(info, Maze, x, y, dir);
      return 0;
    }
    else
      return -1;
    break;
  case DONE:
    return 0;
    break;
  default:
    return -1;
    break;
  }
}

/* Draw View Screen */
void DrawView(Widget w, int width, int hight, void *info)
{
  XPoint points[4];
  int x1 = width / 4;
  int x2 = width - width / 4;
  int xc = width / 2;
  int y1 = hight / 4;
  int y2 = hight - hight / 4;
  int yc = hight / 2;

  GetViewInfo((ViewInfo *)info, Maze, PosX, PosY, Dir);

#if 0
  fprintf(stderr, "(%d,%d: %d) %d %d %d %d\n",
	  PosX, PosY, Dir, View.left, View.front, View.right, View.current);
#endif

  SetDrawArea(w);
  ClearDrawArea();

  points[0].x = 0;
  points[0].y = 0;
  points[1].x = x1;
  points[1].y = y1;
  points[2].x = x2;
  points[2].y = y1;
  points[3].x = width;
  points[3].y = 0;
  SetColor(BLUE);
  DrawFilledPolygon(points, 4);

  points[0].y = hight;
  points[1].y = y2;
  points[2].y = y2;
  points[3].y = hight;
  switch(((ViewInfo *)info)->current) {
  case GOAL:
    SetColor(RED);
    break;
  case START:
    SetColor(GREEN);
    break;
  default:
    SetColor(BLUE);
    break;
  }
  DrawFilledPolygon(points, 4);

  switch(((ViewInfo *)info)->left) {
  case WALL:
    points[0].x = 0;
    points[0].y = 0;
    points[1].x = x1;
    points[1].y = y1;
    points[2].x = x1;
    points[2].y = y2;
    points[3].x = 0;
    points[3].y = hight;
    SetColor(WHITE);
    DrawFilledPolygon(points, 4);
    break;

  case ROUTE:
    points[0].x = 0;
    points[0].y = 0;
    points[1].x = x1;
    points[1].y = y1;
    points[2].x = 0;
    points[2].y = y1;
    SetColor(BLUE);
    DrawFilledPolygon(points, 3);

    SetColor(WHITE);
    DrawFilledBox(0, y1, width / 4, hight / 2);

    points[0].x = 0;
    points[0].y = hight;
    points[1].x = x1;
    points[1].y = y2;
    points[2].x = 0;
    points[2].y = y2;
    SetColor(BLUE);
    DrawFilledPolygon(points, 3);
    break;

  case GOAL:
    points[0].x = 0;
    points[0].y = 0;
    points[1].x = x1;
    points[1].y = y1;
    points[2].x = 0;
    points[2].y = y1;
    SetColor(BLUE);
    DrawFilledPolygon(points, 3);

    SetColor(WHITE);
    DrawFilledBox(0, y1, width / 4, hight / 2);

    points[0].x = 0;
    points[0].y = hight;
    points[1].x = x1;
    points[1].y = y2;
    points[2].x = 0;
    points[2].y = y2;
    SetColor(RED);
    DrawFilledPolygon(points, 3);
    break;

  case START:
    points[0].x = 0;
    points[0].y = 0;
    points[1].x = x1;
    points[1].y = y1;
    points[2].x = 0;
    points[2].y = y1;
    SetColor(BLUE);
    DrawFilledPolygon(points, 3);

    SetColor(WHITE);
    DrawFilledBox(0, y1, width / 4, hight / 2);

    points[0].x = 0;
    points[0].y = hight;
    points[1].x = x1;
    points[1].y = y2;
    points[2].x = 0;
    points[2].y = y2;
    SetColor(GREEN);
    DrawFilledPolygon(points, 3);
    break;
  }

  switch(((ViewInfo *)info)->front) {
  case WALL:
    SetColor(WHITE);
    DrawFilledBox(x1, y1,width / 2, hight / 2);
    break;

  case ROUTE:
    points[0].x = x1;
    points[0].y = y1;
    points[1].x = xc;
    points[1].y = yc;
    points[2].x = x1;
    points[2].y = y2;
    SetColor(WHITE);
    DrawFilledPolygon(points, 3);

    points[0].x = x2;
    points[2].x = x2;
    DrawFilledPolygon(points, 3);

    points[0].x = x1;
    points[0].y = y1;
    points[1].x = xc;
    points[1].y = yc;
    points[2].x = x2;
    points[2].y = y1;
    SetColor(BLUE);
    DrawFilledPolygon(points, 3);

    points[0].y = y2;
    points[2].y = y2;
    DrawFilledPolygon(points, 3);
    break;

  case GOAL:
    points[0].x = x1;
    points[0].y = y1;
    points[1].x = xc;
    points[1].y = yc;
    points[2].x = x1;
    points[2].y = y2;
    SetColor(WHITE);
    DrawFilledPolygon(points, 3);

    points[0].x = x2;
    points[2].x = x2;
    DrawFilledPolygon(points, 3);

    points[0].x = x1;
    points[0].y = y1;
    points[1].x = xc;
    points[1].y = yc;
    points[2].x = x2;
    points[2].y = y1;
    SetColor(BLUE);
    DrawFilledPolygon(points, 3);

    points[0].y = y2;
    points[2].y = y2;
    SetColor(RED);
    DrawFilledPolygon(points, 3);
    break;

  case START:
    points[0].x = x1;
    points[0].y = y1;
    points[1].x = xc;
    points[1].y = yc;
    points[2].x = x1;
    points[2].y = y2;
    SetColor(WHITE);
    DrawFilledPolygon(points, 3);

    points[0].x = x2;
    points[2].x = x2;
    DrawFilledPolygon(points, 3);

    points[0].x = x1;
    points[0].y = y1;
    points[1].x = xc;
    points[1].y = yc;
    points[2].x = x2;
    points[2].y = y1;
    SetColor(BLUE);
    DrawFilledPolygon(points, 3);

    points[0].y = y2;
    points[2].y = y2;
    SetColor(GREEN);
    DrawFilledPolygon(points, 3);
    break;
  }

  switch(((ViewInfo *)info)->right) {
  case WALL:
    points[0].x = width;
    points[0].y = 0;
    points[1].x = x2;
    points[1].y = y1;
    points[2].x = x2;
    points[2].y = y2;
    points[3].x = width;
    points[3].y = hight;
    SetColor(WHITE);
    DrawFilledPolygon(points, 4);
    break;

  case ROUTE:
    points[0].x = width;
    points[0].y = 0;
    points[1].x = x2;
    points[1].y = y1;
    points[2].x = width;
    points[2].y = y1;
    SetColor(BLUE);
    DrawFilledPolygon(points, 3);

    SetColor(WHITE);
    DrawFilledBox(0, y1, width / 4, hight / 2);

    points[0].x = width;
    points[0].y = hight;
    points[1].x = x2;
    points[1].y = y2;
    points[2].x = width;
    points[2].y = y2;
    SetColor(BLUE);
    DrawFilledPolygon(points, 3);
    break;

  case GOAL:
    points[0].x = width;
    points[0].y = 0;
    points[1].x = x2;
    points[1].y = y1;
    points[2].x = width;
    points[2].y = y1;
    SetColor(BLUE);
    DrawFilledPolygon(points, 3);

    SetColor(WHITE);
    DrawFilledBox(0, y1, width / 4, hight / 2);

    points[0].x = width;
    points[0].y = hight;
    points[1].x = x2;
    points[1].y = y2;
    points[2].x = width;
    points[2].y = y2;
    SetColor(RED);
    DrawFilledPolygon(points, 3);
    break;

  case START:
    points[0].x = width;
    points[0].y = 0;
    points[1].x = x2;
    points[1].y = y1;
    points[2].x = width;
    points[2].y = y1;
    SetColor(BLUE);
    DrawFilledPolygon(points, 3);

    SetColor(WHITE);
    DrawFilledBox(0, y1, width / 4, hight / 2);

    points[0].x = width;
    points[0].y = hight;
    points[1].x = x2;
    points[1].y = y2;
    points[2].x = width;
    points[2].y = y2;
    SetColor(GREEN);
    DrawFilledPolygon(points, 3);
    break;
  }
}

/* Draw Map Screen */
void DrawMap(Widget w, int width, int hight, void *data)
{
  int x, y;
  int x1, y1;
  int w1 = width / HSize;
  int h1 = hight / VSize;
  int w2 = width / HSize / 2;
  int h2 = hight / VSize / 2;
  XPoint points[3];

  SetDrawArea(w);
  ClearDrawArea();
  for (x = 0; x < HSize; x++) {
    x1 = (width * x) / HSize;
    for (y = 0; y < VSize; y++) {
      y1 = (hight * y) / VSize;

      switch(SysmtemMap[x][y]) {
      case ROUTE:
	SetColor(BLUE);
	break;
      case WALL:
	SetColor(WHITE);
	break;
      case GOAL:
	SetColor(RED);
	break;
      case START:
	SetColor(GREEN);
	break;
      default:
	SetColor(WHITE);
	break;
      }
      DrawFilledBox(x1, y1, w1, h1);

      if ((x == PosX) && (y == PosY)) {
	switch(Dir) {
	case 0:
	  points[0].x = x1 + w2;
	  points[0].y = y1;
	  points[1].x = x1 + w1;
	  points[1].y = y1 + h1;
	  points[2].x = x1;
	  points[2].y = y1 + h1;
	  break;
	case 1:
	  points[0].x = x1 + w1;
	  points[0].y = y1 + h2;
	  points[1].x = x1;
	  points[1].y = y1 + h1;
	  points[2].x = x1;
	  points[2].y = y1;
	  break;
	case 2:
	  points[0].x = x1 + w2;
	  points[0].y = y1 + h1;
	  points[1].x = x1 + w1;
	  points[1].y = y1;
	  points[2].x = x1;
	  points[2].y = y1;
	  break;
	case 3:
	  points[0].x = x1;
	  points[0].y = y1 + h2;
	  points[1].x = x1 + w1;
	  points[1].y = y1;
	  points[2].x = x1 + w1;
	  points[2].y = y1 + h1;
	  break;
	}
	SetColor(BLACK);
	DrawFilledPolygon(points, 3);
      }
    }
  }
}

/* Update Map Screen */
void DrawMapClean(Widget w, int width, int hight, void *data)
{
  int x1, y1;
  int w1 = width / HSize;
  int h1 = hight / VSize;

  SetDrawArea(w);
  x1 = (width * PosX) / HSize;
  y1 = (hight * PosY) / VSize;
  switch(SysmtemMap[PosX][PosY]) {
  case ROUTE:
    SetColor(BLUE);
    break;
  case WALL:
    SetColor(WHITE);
    break;
  case GOAL:
    SetColor(RED);
    break;
  case START:
    SetColor(GREEN);
    break;
  default:
    SetColor(BLACK);
    break;
  }
  DrawFilledBox(x1, y1, w1, h1);
#if 0
  fprintf(stderr, "(%d,%d:%d,%d:%d,%d)\n",
	  x1, y1, width, hight, w1, h1);
#endif
}

/* Update Map Screen */
void DrawMapUpdate(Widget w, int width, int hight, void *data)
{
  int x1, y1;
  int w1 = width / HSize;
  int h1 = hight / VSize;
  int w2 = width / HSize / 2;
  int h2 = hight / VSize / 2;
  XPoint points[3];

  SetDrawArea(w);
  x1 = (width * PosX) / HSize;
  y1 = (hight * PosY) / VSize;
  switch(SysmtemMap[PosX][PosY]) {
  case ROUTE:
    SetColor(BLUE);
    break;
  case WALL:
    SetColor(WHITE);
    break;
  case GOAL:
    SetColor(RED);
    break;
  case START:
    SetColor(GREEN);
    break;
  default:
    SetColor(BLACK);
    break;
  }
  DrawFilledBox(x1, y1, w1, h1);
  switch(Dir) {
  case 0:
    points[0].x = x1 + w2;
    points[0].y = y1;
    points[1].x = x1 + w1;
    points[1].y = y1 + h1;
    points[2].x = x1;
    points[2].y = y1 + h1;
    break;
  case 1:
    points[0].x = x1 + w1;
    points[0].y = y1 + h2;
    points[1].x = x1;
    points[1].y = y1 + h1;
    points[2].x = x1;
    points[2].y = y1;
    break;
  case 2:
    points[0].x = x1 + w2;
    points[0].y = y1 + h1;
    points[1].x = x1 + w1;
    points[1].y = y1;
    points[2].x = x1;
    points[2].y = y1;
    break;
  case 3:
    points[0].x = x1;
    points[0].y = y1 + h2;
    points[1].x = x1 + w1;
    points[1].y = y1;
    points[2].x = x1 + w1;
    points[2].y = y1 + h1;
    break;
  }
  SetColor(BLACK);
  DrawFilledPolygon(points, 3);
#if 0
  fprintf(stderr, "(%d,%d:%d,%d:%d,%d)\n",
	  x1, y1, width, hight, w1, h1);
#endif
}

/* Quit */
void Quit(Widget w, void *data)
{
  exit(0);
}

/* Turn Right */
void TurnRight(Widget w, void *data)
{
  if (++Dir > 3)
    Dir -= 4;
  DrawView(ViewScreen, ViewWidth, ViewHight, &View);
  DrawMapUpdate(MapScreen, MapWidth, MapHight, NULL);
}

/* Turn Left */
void TurnLeft(Widget w, void *data)
{
  if (--Dir < 0)
    Dir += 4;
  DrawView(ViewScreen, ViewWidth, ViewHight, &View);
  DrawMapUpdate(MapScreen, MapWidth, MapHight, NULL);
}

/* Forward */
void Forward(Widget w, void *data)
{
  if (View.front != WALL) {
    DrawMapClean(MapScreen, MapWidth, MapHight, NULL);
    PosX += DirX[Dir];
    PosY += DirY[Dir];
    DrawView(ViewScreen, ViewWidth, ViewHight, &View);
    SysmtemMap[PosX][PosY] = View.current;
    DrawMapUpdate(MapScreen, MapWidth, MapHight, NULL);

    switch(Stat) {
    case STAT_STEP1:
      Step1Count++;
      sprintf(StepString, "Count: %d", Step1Count);
      SetLabel(Count1Label, StepString);
      break;
    case  STAT_STEP2:
      Step2Count++;
      sprintf(StepString, "Count: %d", Step2Count);
      SetLabel(Count2Label, StepString);
      break;
    }
  }
}


/* Step1 Single-Step */
void Step1Run(void *data)
{
  switch(Step1FindNext()) {
  case TURN_LEFT:
    TurnLeft(ViewScreen, NULL);
    break;
  case GO_LEFT:
    TurnLeft(ViewScreen, NULL);
    Forward(ViewScreen, NULL);
    break;
  case GO_FORWARD:
    Forward(ViewScreen, NULL);
    break;
  case GO_RIGHT:
    TurnRight(ViewScreen, NULL);
    Forward(ViewScreen, NULL);
    break;
  case TURN_RIGHT:
    TurnRight(ViewScreen, NULL);
    break;
  case GO_BACK:
    TurnLeft(ViewScreen, NULL);
    TurnLeft(ViewScreen, NULL);
    Forward(ViewScreen, NULL);
    break;
  case DONE:
    AutoRun = 0;
    Stat = STAT_SEARCH;
    SetLabel(StatLabel, "Stat: Search");
    SetLabel(Auto1Button, "Auto");
    break;
  default:
    break;
  }

  if (AutoRun) {
    AddTimeOut(TIMEOUT, Step1Run, NULL);
  }
}

/* Step1 Single-Step Button */
void Step1(Widget w, void *data)
{
  if (Stat != STAT_STEP1) {
    return;
  }

  Step1Run(NULL);
}

/* Step1 Auto Run */
void Auto1(Widget w, void *data)
{
  if (Stat != STAT_STEP1) {
    return;
  }

  AutoRun = 1 - AutoRun;

  if (AutoRun) {
    SetLabel(w, "Stop");
    AddTimeOut(TIMEOUT, Step1Run, NULL);
  }
  else {
    SetLabel(w, "Auto");
  }
}

/* Search */
void Search(Widget w, void *data)
{
  if (Stat != STAT_SEARCH) {
    return;
  }

  SearchPath();

  Stat = STAT_STEP2;
  SetLabel(StatLabel, "Stat: Step2  ");
  Dir = 3;
  PosX = StartX;
  PosY = StartY;
  DrawView(ViewScreen, ViewWidth, ViewHight, &View);
  DrawMap(MapScreen, MapWidth, MapHight, NULL);
}


/* Step2 Single-Step */
void Step2Run(void *data)
{
  switch(Step2FindNext()) {
  case TURN_LEFT:
    TurnLeft(ViewScreen, NULL);
    break;
  case GO_LEFT:
    TurnLeft(ViewScreen, NULL);
    Forward(ViewScreen, NULL);
    break;
  case GO_FORWARD:
    Forward(ViewScreen, NULL);
    break;
  case GO_RIGHT:
    TurnRight(ViewScreen, NULL);
    Forward(ViewScreen, NULL);
    break;
  case TURN_RIGHT:
    TurnRight(ViewScreen, NULL);
    break;
  case GO_BACK:
    TurnLeft(ViewScreen, NULL);
    TurnLeft(ViewScreen, NULL);
    Forward(ViewScreen, NULL);
    break;
  case DONE:
    AutoRun = 0;
    Stat = STAT_DONE;
    SetLabel(StatLabel, "Stat: Done   ");
    SetLabel(Auto2Button, "Auto");
    break;
  default:
    break;
  }

  if (AutoRun) {
    AddTimeOut(TIMEOUT, Step2Run, NULL);
  }
}

/* Step2 Single-Step Button */
void Step2(Widget w, void *data)
{
  if (Stat != STAT_STEP2) {
    return;
  }

  Step2Run(NULL);
}

/* Step2 Auto Run */
void Auto2(Widget w, void *data)
{
  if (Stat != STAT_STEP2) {
    return;
  }

  AutoRun = 1 - AutoRun;

  if (AutoRun) {
    SetLabel(w, "Stop");
    AddTimeOut(TIMEOUT, Step2Run, NULL);
  }
  else {
    SetLabel(w, "Auto");
  }
}

/* Initialize Screen */
void InitScreen(int argc, char *argv[])
{
  if (OpenDisplay(argc, argv) == 0) {
    fprintf(stderr, "Cannot open display\n");
    exit(1);
  }

  QuitButton = MakeButton("Quit", Quit, NULL);
  TurnLeftButton = MakeButton("Left", TurnLeft, NULL);
  ForwardButton = MakeButton("Forward", Forward, NULL);
  TurnRightButton = MakeButton("Right", TurnRight, NULL);
  StatLabel = MakeLabel("Stat: Step1  ");

  Step1Label = MakeLabel("Step1: ");
  Step1Button = MakeButton("Step", Step1, NULL);
  Auto1Button = MakeButton("Auto", Auto1, NULL);
  Count1Label = MakeLabel("Count: 0    ");

  SearchLabel = MakeLabel("Search: ");
  SearchButton = MakeButton("Search", Search, NULL);

  Step2Label = MakeLabel("Step2: ");
  Step2Button = MakeButton("Step", Step2, NULL);
  Auto2Button = MakeButton("Auto", Auto2, NULL);
  Count2Label = MakeLabel("Count: 0    ");

  ViewScreen = MakeDrawArea(VIEWSIZE, VIEWSIZE, DrawView, &View);
  MapScreen = MakeDrawArea(MAPSIZE, MAPSIZE, DrawMap, NULL);

  SetWidgetPos(TurnLeftButton, PLACE_RIGHT, QuitButton, NO_CARE, NULL);
  SetWidgetPos(ForwardButton, PLACE_RIGHT, TurnLeftButton, NO_CARE, NULL);
  SetWidgetPos(TurnRightButton, PLACE_RIGHT, ForwardButton, NO_CARE, NULL);
  SetWidgetPos(StatLabel, PLACE_RIGHT, TurnRightButton, NO_CARE, NULL);

  SetWidgetPos(Step1Label, PLACE_UNDER, QuitButton, NO_CARE, NULL);
  SetWidgetPos(Step1Button, PLACE_UNDER, QuitButton, PLACE_RIGHT, Step1Label);
  SetWidgetPos(Auto1Button, PLACE_UNDER, QuitButton, PLACE_RIGHT, Step1Button);
  SetWidgetPos(Count1Label, PLACE_UNDER, QuitButton, PLACE_RIGHT, Auto1Button);

  SetWidgetPos(SearchLabel, PLACE_UNDER, QuitButton, PLACE_RIGHT, Count1Label);
  SetWidgetPos(SearchButton,
	       PLACE_UNDER, QuitButton, PLACE_RIGHT, SearchLabel);

  SetWidgetPos(Step2Label, PLACE_UNDER, QuitButton, PLACE_RIGHT, SearchButton);
  SetWidgetPos(Step2Button, PLACE_UNDER, QuitButton, PLACE_RIGHT, Step2Label);
  SetWidgetPos(Auto2Button, PLACE_UNDER, QuitButton, PLACE_RIGHT, Step2Button);
  SetWidgetPos(Count2Label, PLACE_UNDER, QuitButton, PLACE_RIGHT, Auto2Button);

  SetWidgetPos(ViewScreen, PLACE_UNDER, Step1Label, NO_CARE, NULL);
  SetWidgetPos(MapScreen, PLACE_UNDER, Step1Label, PLACE_RIGHT, ViewScreen);

  ShowDisplay();
  GetStandardColors();
  SetDrawArea(ViewScreen);
  GetDrawAreaSize(&ViewWidth, &ViewHight);
  SetDrawArea(MapScreen);
  GetDrawAreaSize(&MapWidth, &MapHight);
}

main(int argc, char *argv[])
{
  if (argc < 2) {
    fprintf(stderr, "Usage: %s FILE\n", argv[0]);
    exit(1);
  }
  Load(argv[1]);
  Show(Maze);

  InitScreen(argc, argv);
  Init();

  MainLoop();
}
